@@ -14,8 +14,7 @@
 #
 # @license: http://www.apache.org/licenses/LICENSE-2.0
 # @author: Kevin Crouse (krcrouse@gmail.com)
-
-import os, os.path
+import os
 import datetime
 import re
 import shutil
@@ -23,10 +22,15 @@ import shutil
 import warnings
 from pprint import pprint
 
-import pickle
-import google.auth, google.oauth2
-import google_auth_oauthlib
-import googleapiclient.discovery # to install this : pip install --upgrade google-api-python-client
+import httplib2
+import apiclient
+
+import oauth2client
+import oauth2client.file
+import oauth2client.tools
+
+import googleapiclient # to install this : pip install --upgrade google-api-python-client
+
 
 #-------------- Notes on Permission Scope
 # Permission Scopes are the Google-designed access scopes for all of the
@@ -60,14 +64,14 @@ DEFAULT_SCOPES = {
 
 class ServiceManager(object):
     '''
-    The ServiceManager is intended to encapsulate all of the Google services for a user. The SM object must be tied to the registered OAUTH secret file, which the user must download from the Google Developer Console (https://console.developers.google.com/apis/credentials). The filename from Google will be "client_secret_{id}.json", but if you want to keep your scripts simple, the ServiceManager default looks for "google_client.json" in the "credentials" directory. Of course, these can be overriden and you can manually point to a client secret file.
+    The ServiceManager is intended to encapsulate all of the Google services for a user. The SM object must be tied to the registered OAUTH secret file, which the user must download from the Google Developer Console (https://console.developers.google.com/apis/credentials). The filename from Google will be "client_secret_{ids}.json", but the ServiceManager default looks for "google_client.json" in the "credentials" directory, but these can be overriden.
 
     Notes:
-        Services are cached by default. So, attempt to access the Sheets service from two different parts of an application will refer to the same service and service manager. This is done so you don't have to keep track of service manager objects in modules that are designed to be independent of each other, and, in most cases, this is the correct model because it ends up functioning like a singleton as only one user is authenticating to Google for a given execution. However, since there are potential needs to reference multiple paths, it can be explicitly instantiated to two different paths, in which you will need to balance multiple users, each connected to a separate service managers.
+        Services are cached so that a reference to a Sheets service from two different parts of an application refer to the same service and service manager. In most cases, it ends up functioning like a singleton as only one user is likely to be authenticating to Google for a given script. But since there are plenty of potential needs to reference multiple paths, it can be explicitly instantiated to two different paths.
     '''
     api_key=None
-    secret_file = 'google_client.json'
-    credential_directory='credentials'
+    oauth_secret_file = 'google_client.json'
+    oauth_credential_directory='credentials'
     oauth_args = []
 
     loaded_services = {}
@@ -88,23 +92,23 @@ class ServiceManager(object):
     def __init__(self,
                  services=None,
                  api_key=None,
-                 secret_file=None,
-                 credential_directory=None,
+                 oauth_secret_file=None,
+                 oauth_credential_directory=None,
                  ):
         """
         Args:
             services (list, optional): A list of services to load up initially.
             api_key (str, optional): Uses the proivded developer API Key instead of OAUTH2 for authentication.
-            secret_file (filename): The filename of the Google Developer OAUTH Credential File. By default, this is set to google_client.json. If no file exists, an exception is thrown on first attempt to authenticate.
-            credential_directory (path): The path to the credential directory where the OATH credential file is and where service authorization tokens should be stored. By default, this is the credentials directory from wherever the program is being run. There are no built-in options to have a default  credential location because any guest would potentially be able to access all the user's Google data.
+            oauth_secret_file (filename): The filename of the Google Developer OAUTH Credential File. By default, this is set to google_client.json. If no file exists, an exception is thrown on first attempt to authenticate.
+            oauth_credential_directory (path): The path to the credential directory where the OATH credential file is and where service authorization tokens should be stored. By default, this is the credentials directory from wherever the program is being run. There are no built-in options to have a default  credential location because any guest would potentially be able to access all the user's Google data.
         """
 
         if api_key:
             self.api_key=api_key
-        if credential_directory:
-            self.credential_directory = credential_directory
-        if secret_file:
-            self.secret_file = secret_file
+        if oauth_credential_directory:
+            self.oauth_credential_directory = oauth_credential_directory
+        if oauth_secret_file:
+            self.oauth_secret_file = oauth_secret_file
         if services:
             self.load_services(services)
 
@@ -138,50 +142,54 @@ class ServiceManager(object):
         else:
             return
 
-    def load_service(self, api_name, api_version=None, permission_scope=None, force_refresh=False):
+    def load_service(self, service_key=None, service_path=None, permission_scope=None, force_refresh=False):
         """ Loads the requested service. If the service has already been loaded for this ServiceManager, return the previously loaded service (unless force_refresh is on). If it has not, delegate to get_credentials() as necessary.
         Args:
-            api_name (str): The Google API name (i.e., 'drive.googleapis.com') or a shorthand version of it (e.g., 'drive').
-            api_version (str, optional): The version of the API to use. If not provided, it will first check the DEFAULT_SCOPES mapping to determine if there is a default API. If not, it will attempt to use the directory service to determine the default API.
-            permission_scope (str): The permission scope that accompanies this service. If the service_key is listed in the DEFAULT_SCOPES, this is optional and the default will be pulled. As fully-qualified permission scopes generally start with "https://www.googleapis.com/auth/", that part is optional and will be added if not present.
+            service_key (str): The service key, which may be a shorthand value ('drive') or a qualified google api path ('drive.googleapis.com'). If not provided, the service key will be pulled from the first part of the service_path string.
+            service_path (str): The fullly qualified service path, which also becomes the filename in the credential directory where the authorization token is stored. If not provided, this is generated from the service key ({service_key}.googleapis.com-python.json)
+            permission_scope (str): The permission scope that accompanies this service. If the service_key is listed in the DEFAULT_SCOPES, this is optional and the default will be pulled.
             force_refresh (bool): Reload the service even if it already has been loaded. Default is False.
         Returns:
             The loaded service.
         """
-        #OLD and NEEDING TO DELETE: The fullly qualified service path, which also becomes the filename in the credential directory where the authorization token is stored. If not provided, this is generated from the service key ({service_key}.googleapis.com-python.json)
 
-        if api_name in DEFAULT_SCOPES:
-            keydata = DEFAULT_SCOPES[api_name]
-            api_name = keydata[0]
-            if not api_version:
-                api_version = keydata[1]
-            if not permission_scope:
-                permission_scope = 'https://www.googleapis.com/auth/' + keydata[2]
-
-        if not api_version:
-            #TODO
-            raise Exception("Pulling the discovery to determine the default version is not yet available")
-
-        service_key = f"{api_name}.{api_version}"
-
-        # Return the already-loaded service if it exists. This provides singleton-ish functionality at a Service:User level.
+        # only actually load if it needs to be
         if service_key in self.loaded_services and not force_refresh:
-            return(self.loaded_services[service_key])
+            return self.loaded_services[service_key]
 
         # either service_key or service_path is required
+        if not service_key:
+            if service_path:
+                match = re.match('(.*)\.googleapis', service_path)
+                service_key = match.group(1)
+            else:
+                raise Exception("load_services() requires either the service_key or the service_path to be provided")
+
+        # determine
+        if service_key in DEFAULT_SCOPES:
+            keydata = DEFAULT_SCOPES[service_key]
+            api_name = keydata[0]
+            version = keydata[1]
+            if not permission_scope:
+                permission_scope = 'https://www.googleapis.com/auth/' + keydata[2]
 
         if self.api_key:
             # api keys should be used only for personal application and development
             # anyone with the api key can access Google services as the developer
-            # TODO: This has not been tested in a while
-            discovery = googleapiclient.discovery.build(api_name, api_version, developerKey=self.api_key)
-        elif self.secret_file:
+            discovery = apiclient.discovery.build(api_name, version, developerKey=self.api_key)
+        elif self.oauth_secret_file:
             # engage in OAuth2 authorization and credentialling
 
-            service_path = f"{api_name}.{api_version}.googleapis.json"
+            if not service_path:
+                service_path = service_key + '.googleapis.com-python.json'
+
+            # convert the shorthand permission scope to the full path
+            if not re.match('https', permission_scope):
+                permission_scope = 'https://www.googleapis.com/auth/' + permission_scope
+
             credentials = self.get_credentials(service_path, permission_scope)
-            #authorizer = credentials.authorize(httplib2.Http())
-            discovery = googleapiclient.discovery.build(api_name, api_version, credentials=credentials) #http=authorizer)
+            authorizer = credentials.authorize(httplib2.Http())
+            discovery = apiclient.discovery.build(api_name, version, http=authorizer)
         else:
             raise Exception("Unable to load_service as no authentication models provided. Please specify an API Key or provide OAuth credentials to the googleapps.service.ServiceManager")
 
@@ -190,60 +198,51 @@ class ServiceManager(object):
         return(self.loaded_services[service_key])
 
 
-    def get_credentials(self, local_credential_file, permission_scope):
+    def get_credentials(self, local_credential_file, permission_scope=None):
         """Gets valid user credentials - in most cases this is not required to be called directly and will be called from load_service().
 
         If nothing has been stored, or if the stored credentials are invalid, AND if application secret and scope are provided, the OAuth2 flow is completed to obtain the new credentials. This requires the user to authorize access via a web browser.
 
         Args:
-            local_credential_file (path): the service-specific credential file, which is an authorization for this program/user to access these services. This file must exist in the credential_directory defined for this service object, or it will attempt to be fetched following user authorization. In the most common case where get_credentials is called by load_service(), this is determined in that function.
+            local_credential_file (path): the service-specific credential file, which is an authorization for this program/user to access these services. This file must exist in the oauth_credential_directory defined for this service object, or it will attempt to be fetched following user authorization. In the most common case where get_credentials is called by load_service(), this is determined in that function.
             permission_score (str): the permissions necessary for the application and that a remote call to authorize will request. The user may provide just the scope name (drive.readonly) or the full path to the permission scope (i.e. https://www.googleapis.com/auth/drive.readonly)
 
         Returns:
             Credentials, the obtained credential.
         """
+
         # auto create directory
-        if not os.path.exists(self.credential_directory):
-            os.makedirs(self.credential_directory)
+        if not os.path.exists(self.oauth_credential_directory):
+            os.makedirs(self.oauth_credential_directory)
 
-        # convert the shorthand permission scope to the full path
-        if not re.match('https', permission_scope):
-            permission_scope = 'https://www.googleapis.com/auth/' + permission_scope
+        credential_path = os.path.join(self.oauth_credential_directory, local_credential_file)
+        store = oauth2client.file.Storage(credential_path)
+        credentials = store.get()
 
+        if not credentials or credentials.invalid :
+            if not permission_scope:
+                raise Exception("No permission scope provided to get_credentials() called for " + local_credential_file + ", which does not exist or is not valid and needs to be authenticated.")
 
-        # determine ifr the credentials have already been authorized and are active
-        credential_path = os.path.join(self.credential_directory, local_credential_file)
-        if os.path.exists(credential_path):
-            credentials = google.oauth2.credentials.Credentials.from_authorized_user_file(credential_path, permission_scope)
-        else:
-            credentials = None
-        if credentials and credentials.valid:
-            if credentials.expired:
-                if credentials.refresh_token:
-                    # creds exist and arre valid, but are expired - so just refresh
-                    credentials.refresh(google.auth.transport.requests.Request())
-                    # Save the credentials for the next run
-                    with open(credential_path, 'w') as token:
-                        token.write(credentials.to_json())
-                    return(credentials)
+            # identify the path to the credential's secret file.
+            oauth_secret_path = os.path.join(self.oauth_credential_directory, self.oauth_secret_file)
+
+            # a very detailed error message.
+            if not os.path.exists(oauth_secret_path):
+                raise FileNotFoundError(f"Could not find an OAuth secret file at {oauth_secret_path}, which is required for ServiceManager.get_credentials() to attempt to complete the authorization permission for scope {permission_scope} (the scope-level permission should be stored in {local_credential_file}, but it does not exist or is no longer valid and needs to be authenticated). \n\nYou will need to indicate the correct path to your OAuth Credential JSON secret file or download a new copy from the Google Developers Console located at https://console.developers.google.com/apis/credentials/ before trying again.")
+
+            # All set.
+            # Do OAuth authentication
+            if not re.match('https', permission_scope):
+                permission_scope = 'https://www.googleapis.com/auth/' + permission_scope
+
+            # to date, I've never set any oauth args.
+            flags = oauth2client.tools.argparser.parse_args(args=self.oauth_args)
+            flow = oauth2client.client.flow_from_clientsecrets(
+                oauth_secret_path,
+                permission_scope
+            )
+            if flags:
+                credentials = oauth2client.tools.run_flow(flow, store, flags)
             else:
-                return(credentials)
-
-        # no credientials, or they are invalid!
-        # -- get new credentials
-        # identify the path to the credential's secret file.
-        secret_path = os.path.join(self.credential_directory, self.secret_file)
-
-        # a very detailed error message.
-        if not os.path.exists(secret_path):
-            raise FileNotFoundError(f"Could not find an OAuth secret file at {secret_path}, which is required for ServiceManager.get_credentials() to attempt to complete the authorization permission for scope {permission_scope} (the scope-level permission should be stored in {local_credential_file}, but it does not exist or is no longer valid and needs to be authenticated). \n\nYou will need to indicate the correct path to your OAuth Credential JSON secret file or download a new copy from the Google Developers Console located at https://console.developers.google.com/apis/credentials/ before trying again.")
-
-        # All set authorize flow.
-        flow = google_auth_oauthlib.flow.InstalledAppFlow.from_client_secrets_file(secret_path, permission_scope)
-        credentials = flow.run_local_server(port=0)
-
-        # Save the credentials for the next run
-        with open(credential_path, 'w') as token:
-            token.write(credentials.to_json())
-
+                credentials = oauth2client.tools.run(flow, store)
         return(credentials)
